from collections import defaultdict
from typing import Any, Iterator

from ..base import BaseSolver


def in_correct_order(update: list[int], requirements: dict[int, set[int]]) -> bool:
    return all(
        not any(value_2 in requirements[value] for value_2 in update[i_value:])
        for i_value, value in enumerate(update)
    )


def to_correct_order(
    update: list[int],
    requirements: dict[int, set[int]],
    max_update_length: int | None = None,
) -> list[int]:
    # copy requirements to update
    requirements = {
        value: {predecessor for predecessor in predecessors if predecessor in update}
        for value, predecessors in requirements.items()
        if value in update
    }

    max_update_length = max_update_length or len(update)

    update = []
    while requirements and len(update) < max_update_length:
        value = next(
            value for value, predecessors in requirements.items() if not predecessors
        )

        update.append(value)
        del requirements[value]

        for predecessors in requirements.values():
            if value in predecessors:
                predecessors.remove(value)

    return update


class Solver(BaseSolver):
    def solve(self, input: str) -> Iterator[Any]:
        part1, part2 = input.split("\n\n")

        requirements: dict[int, set[int]] = defaultdict(set)
        for line in part1.splitlines():
            v1, v2 = line.split("|")
            requirements[int(v2)].add(int(v1))

        updates = [list(map(int, line.split(","))) for line in part2.splitlines()]

        yield sum(
            update[len(update) // 2]
            for update in updates
            if in_correct_order(update, requirements)
        )

        yield sum(
            to_correct_order(update, requirements, len(update) // 2 + 1)[-1]
            for update in updates
            if not in_correct_order(update, requirements)
        )
